# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import importlib.resources
import json
import os
import textwrap
import types

import certifi
import pkg_resources
import pykube
import yaml
from magnum import objects as magnum_objects
from magnum.common import context, neutron
from magnum.common.cert_manager import cert_manager
from magnum.common.x509 import operations as x509
from magnum.conductor.handlers.common import cert_manager as cert_manager_handlers
from oslo_config import cfg
from oslo_serialization import base64
from oslo_utils import encodeutils

from magnum_cluster_api import (
    clients,
    helm,
    image_utils,
    images,
    json_patches,
    objects,
    utils,
)
from magnum_cluster_api.integrations import cinder, cloud_provider, manila

CONF = cfg.CONF
CALICO_TAG = "v3.24.2"

CLUSTER_CLASS_VERSION = pkg_resources.require("magnum_cluster_api")[0].version
CLUSTER_CLASS_NAME = f"magnum-v{CLUSTER_CLASS_VERSION}"
CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT = "300s"  # seconds

PLACEHOLDER = "PLACEHOLDER"

AUTOSCALE_ANNOTATION_MIN = "cluster.x-k8s.io/cluster-api-autoscaler-node-group-min-size"
AUTOSCALE_ANNOTATION_MAX = "cluster.x-k8s.io/cluster-api-autoscaler-node-group-max-size"


class ClusterAutoscalerHelmRelease:
    def __init__(self, api, cluster) -> None:
        self.cluster = cluster

    @property
    def apply(self):
        image = images.get_cluster_autoscaler_image(
            utils.get_kube_tag(self.cluster),
        )
        image_repo, image_tag = image.split(":", 1)

        return helm.UpgradeReleaseCommand(
            namespace="magnum-system",
            release_name=self.cluster.stack_id,
            chart_ref=os.path.join(
                pkg_resources.resource_filename("magnum_cluster_api", "charts"),
                "cluster-autoscaler/",
            ),
            values={
                "fullnameOverride": f"{self.cluster.stack_id}-autoscaler",
                "cloudProvider": "clusterapi",
                "clusterAPIMode": "kubeconfig-incluster",
                "clusterAPIKubeconfigSecret": f"{self.cluster.stack_id}-kubeconfig",
                "autoDiscovery": {
                    "clusterName": self.cluster.stack_id,
                },
                "image": {
                    "repository": image_repo,
                    "tag": image_tag,
                },
                "nodeSelector": {
                    "openstack-control-plane": "enabled",
                },
            },
        )

    @property
    def delete(self):
        return helm.DeleteReleaseCommand(
            namespace="magnum-system",
            release_name=self.cluster.stack_id,
            skip_missing=True,
        )


class Base:
    def __init__(self, api: pykube.HTTPClient):
        self.api = api

    def apply(self) -> None:
        resource = self.get_object()
        resp = resource.api.patch(
            **resource.api_kwargs(
                headers={
                    "Content-Type": "application/apply-patch+yaml",
                },
                params={
                    "fieldManager": "atmosphere-operator",
                    "force": True,
                },
                data=json.dumps(resource.obj),
            )
        )

        resource.api.raise_for_status(resp)
        resource.set_obj(resp.json())

    def delete(self) -> None:
        resource = self.get_object()
        resource.delete()


class Namespace(Base):
    def get_object(self) -> pykube.Namespace:
        return pykube.Namespace(
            self.api,
            {
                "apiVersion": pykube.Namespace.version,
                "kind": pykube.Namespace.kind,
                "metadata": {
                    "name": "magnum-system",
                },
            },
        )


class ClusterBase(Base):
    def __init__(self, api: pykube.HTTPClient, cluster: magnum_objects.Cluster):
        super().__init__(api)
        self.cluster = cluster

    @property
    def labels(self) -> dict:
        return {
            "cluster-uuid": self.cluster.uuid,
        }


class ClusterResourcesConfigMap(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
    ):
        self.context = context
        self.api = api
        self.cluster = cluster

    def get_object(self) -> pykube.ConfigMap:
        manifests_path = pkg_resources.resource_filename(
            "magnum_cluster_api", "manifests"
        )
        calico_version = self.cluster.labels.get("calico_tag", CALICO_TAG)

        repository = utils.get_cluster_container_infra_prefix(self.cluster)

        osc = clients.get_openstack_api(self.context)

        data = {
            **{
                os.path.basename(manifest): image_utils.update_manifest_images(
                    self.cluster.uuid,
                    manifest,
                    repository=repository,
                    replacements=[
                        (
                            "docker.io/k8scloudprovider/openstack-cloud-controller-manager:latest",
                            cloud_provider.get_image(self.cluster),
                        ),
                    ],
                )
                for manifest in glob.glob(os.path.join(manifests_path, "ccm/*.yaml"))
            },
            **{
                "calico.yml": image_utils.update_manifest_images(
                    self.cluster.uuid,
                    os.path.join(manifests_path, f"calico/{calico_version}.yaml"),
                    repository=repository,
                )
            },
        }

        if cinder.is_enabled(self.cluster):
            volume_types = osc.cinder().volume_types.list()
            default_volume_type = osc.cinder().volume_types.default()
            data = {
                **data,
                **{
                    os.path.basename(manifest): image_utils.update_manifest_images(
                        self.cluster.uuid,
                        manifest,
                        repository=repository,
                        replacements=[
                            (
                                "docker.io/k8scloudprovider/cinder-csi-plugin:latest",
                                cinder.get_image(self.cluster),
                            ),
                        ],
                    )
                    for manifest in glob.glob(
                        os.path.join(manifests_path, "cinder-csi/*.yaml")
                    )
                },
                **{
                    f"storageclass-block-{vt.name}.yaml": yaml.dump(
                        {
                            "apiVersion": objects.StorageClass.version,
                            "allowVolumeExpansion": True,
                            "kind": objects.StorageClass.kind,
                            "metadata": {
                                "annotations": (
                                    {
                                        "storageclass.kubernetes.io/is-default-class": "true"
                                    }
                                    if default_volume_type.name == vt.name
                                    else {}
                                ),
                                "name": "block-%s" % utils.convert_to_rfc1123(vt.name),
                            },
                            "provisioner": "cinder.csi.openstack.org",
                            "parameters": {
                                "type": utils.convert_to_rfc1123(vt.name),
                            },
                            "reclaimPolicy": "Delete",
                            "volumeBindingMode": "Immediate",
                        }
                    )
                    for vt in volume_types
                    if vt.name != "__DEFAULT__"
                },
            }

        if manila.is_enabled(self.cluster):
            share_types = osc.manila().share_types.list()
            share_network_id = self.cluster.labels.get("manila_csi_share_network_id")
            data = {
                **data,
                **{
                    "manila-csi-secret.yaml": yaml.dump(
                        {
                            "apiVersion": pykube.Secret.version,
                            "kind": pykube.Secret.kind,
                            "metadata": {
                                "name": "csi-manila-secrets",
                                "namespace": "kube-system",
                            },
                            "stringData": utils.generate_manila_csi_cloud_config(
                                self.context,
                                self.api,
                                self.cluster,
                            ),
                        },
                    )
                },
                **{
                    os.path.basename(manifest): image_utils.update_manifest_images(
                        self.cluster.uuid,
                        manifest,
                        repository=repository,
                    )
                    for manifest in glob.glob(
                        os.path.join(manifests_path, "nfs-csi/*.yaml")
                    )
                },
                **{
                    os.path.basename(manifest): image_utils.update_manifest_images(
                        self.cluster.uuid,
                        manifest,
                        repository=repository,
                        replacements=[
                            (
                                "registry.k8s.io/provider-os/manila-csi-plugin:latest",
                                manila.get_image(self.cluster),
                            ),
                        ],
                    )
                    for manifest in glob.glob(
                        os.path.join(manifests_path, "manila-csi/*.yaml")
                    )
                },
            }
            # NOTE: We only create StorageClasses if share_network_id specified.
            if share_network_id:
                data = {
                    **data,
                    **{
                        f"storageclass-share-{st.name}.yaml": yaml.dump(
                            {
                                "apiVersion": objects.StorageClass.version,
                                "allowVolumeExpansion": True,
                                "kind": objects.StorageClass.kind,
                                "metadata": {
                                    "name": "share-%s"
                                    % utils.convert_to_rfc1123(st.name),
                                },
                                "provisioner": "manila.csi.openstack.org",
                                "parameters": {
                                    "type": st.name,
                                    "shareNetworkID": share_network_id,
                                    "csi.storage.k8s.io/provisioner-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/provisioner-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/controller-expand-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/controller-expand-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/node-stage-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/node-stage-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/node-publish-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/node-publish-secret-namespace": "kube-system",
                                },
                                "reclaimPolicy": "Delete",
                                "volumeBindingMode": "Immediate",
                            }
                        )
                        for st in share_types
                    },
                }

        if utils.get_cluster_label_as_bool(self.cluster, "keystone_auth_enabled", True):
            auth_url = osc.url_for(
                service_type="identity",
                interface=CONF.capi_client.endpoint_type.replace("URL", ""),
            )
            data = {
                **data,
                **{
                    "keystone-auth.yaml": helm.TemplateReleaseCommand(
                        namespace="kube-system",
                        release_name="k8s-keystone-auth",
                        chart_ref=os.path.join(
                            pkg_resources.resource_filename(
                                "magnum_cluster_api", "charts"
                            ),
                            "k8s-keystone-auth/",
                        ),
                        values={
                            "conf": {
                                "auth_url": auth_url
                                + ("" if auth_url.endswith("/v3") else "/v3"),
                                "ca_file": utils.get_cloud_ca_cert(),
                                "policy": utils.get_keystone_auth_default_policy(
                                    self.cluster
                                ),
                            },
                        },
                    )(repository=repository)
                },
            }

        return pykube.ConfigMap(
            self.api,
            {
                "apiVersion": pykube.ConfigMap.version,
                "kind": pykube.ConfigMap.kind,
                "metadata": {
                    "name": self.cluster.uuid,
                    "namespace": "magnum-system",
                },
                "data": data,
            },
        )

    def get_or_none(self) -> objects.Cluster:
        return pykube.ConfigMap.objects(
            self.api, namespace="magnum-system"
        ).get_or_none(name=self.cluster.uuid)

    def delete(self):
        cr_cm = self.get_or_none()
        if cr_cm:
            cr_cm.delete()


class ClusterResourceSet(ClusterBase):
    def get_object(self) -> objects.ClusterResourceSet:
        return objects.ClusterResourceSet(
            self.api,
            {
                "apiVersion": objects.ClusterResourceSet.version,
                "kind": objects.ClusterResourceSet.kind,
                "metadata": {
                    "name": self.cluster.uuid,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "clusterSelector": {
                        "matchLabels": {
                            "cluster-uuid": self.cluster.uuid,
                        },
                    },
                    "resources": [
                        {
                            "name": self.cluster.uuid,
                            "kind": "ConfigMap",
                        },
                    ],
                },
            },
        )


class CertificateAuthoritySecret(ClusterBase):
    def __init__(
        self, context: context.RequestContext, api: pykube.HTTPClient, cluster: any
    ):
        super().__init__(api, cluster)
        self.context = context

    def delete(self) -> None:
        resource = self.get_or_none()
        if resource:
            resource.delete()

    def get_or_none(self) -> pykube.Secret:
        return pykube.Secret.objects(self.api, namespace="magnum-system").get_or_none(
            name=f"{self.cluster.stack_id}-{self.CERT}"
        )

    def get_certificate(self) -> cert_manager.Cert:
        raise NotImplementedError()

    def get_object(self) -> pykube.Secret:
        cert_ref = getattr(self.cluster, self.REF)
        if cert_ref is None:
            raise Exception("Certificate for %s doesn't exist." % self.REF)
        ca_cert = self.get_certificate()

        return pykube.Secret(
            self.api,
            {
                "apiVersion": pykube.Secret.version,
                "kind": pykube.Secret.kind,
                "type": "cluster.x-k8s.io/secret",
                "metadata": {
                    "name": f"{self.cluster.stack_id}-{self.CERT}",
                    "namespace": "magnum-system",
                    "labels": {
                        "cluster.x-k8s.io/cluster-name": f"{self.cluster.stack_id}",
                    },
                },
                "stringData": {
                    "tls.crt": encodeutils.safe_decode(ca_cert.get_certificate()),
                    "tls.key": encodeutils.safe_decode(
                        x509.decrypt_key(
                            ca_cert.get_private_key(),
                            ca_cert.get_private_key_passphrase(),
                        )
                    ),
                },
            },
        )


class ApiCertificateAuthoritySecret(CertificateAuthoritySecret):
    CERT = "ca"
    REF = "ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "kubernetes"
        )


class EtcdCertificateAuthoritySecret(CertificateAuthoritySecret):
    CERT = "etcd"
    REF = "etcd_ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "etcd"
        )


class FrontProxyCertificateAuthoritySecret(CertificateAuthoritySecret):
    CERT = "proxy"
    REF = "front_proxy_ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "front-proxy"
        )


class ServiceAccountCertificateAuthoritySecret(CertificateAuthoritySecret):
    CERT = "sa"
    REF = "magnum_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_magnum_cert(self.cluster, self.context)


class CloudConfigSecret(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: pykube.HTTPClient,
        cluster: any,
        region_name: str = None,
        credential: any = types.SimpleNamespace(id=None, secret=None),
    ):
        super().__init__(api, cluster)
        self.context = context
        osc = clients.get_openstack_api(self.context)
        self.auth_url = osc.url_for(
            service_type="identity",
            interface=CONF.capi_client.endpoint_type.replace("URL", ""),
        )
        self.region_name = region_name
        self.credential = credential

    def get_object(self) -> pykube.Secret:
        ca_certificate = utils.get_capi_client_ca_cert()

        return pykube.Secret(
            self.api,
            {
                "apiVersion": pykube.Secret.version,
                "kind": pykube.Secret.kind,
                "metadata": {
                    "name": utils.get_cluster_api_cloud_config_secret_name(
                        self.cluster
                    ),
                    "namespace": "magnum-system",
                    "labels": self.labels,
                },
                "stringData": {
                    "cacert": (
                        ca_certificate
                        if ca_certificate
                        else open(certifi.where(), "r").read()
                    ),
                    "clouds.yaml": yaml.dump(
                        {
                            "clouds": {
                                "default": {
                                    "region_name": self.region_name,
                                    "endpoint_type": CONF.capi_client.endpoint_type.replace(
                                        "URL", ""
                                    ),
                                    "identity_api_version": 3,
                                    "verify": not CONF.capi_client.insecure,
                                    "auth": {
                                        "auth_url": self.auth_url,
                                        "application_credential_id": self.credential.id,
                                        "application_credential_secret": self.credential.secret,
                                    },
                                }
                            }
                        }
                    ),
                },
            },
        )


class KubeadmControlPlaneTemplate(Base):
    def get_object(self) -> objects.KubeadmControlPlaneTemplate:
        manifests_path = pkg_resources.resource_filename(
            "magnum_cluster_api", "manifests"
        )
        audit_policy = open(os.path.join(manifests_path, "audit/policy.yaml")).read()
        keystone_auth_webhook = open(
            os.path.join(manifests_path, "keystone-auth/webhook.yaml")
        ).read()

        return objects.KubeadmControlPlaneTemplate(
            self.api,
            {
                "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                "kind": objects.KubeadmControlPlaneTemplate.kind,
                "metadata": {
                    "name": CLUSTER_CLASS_NAME,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "template": {
                        "spec": {
                            "rolloutBefore": {
                                "certificatesExpiryDays": 21,
                            },
                            "kubeadmConfigSpec": {
                                "clusterConfiguration": {
                                    "apiServer": {
                                        "extraArgs": {
                                            "cloud-provider": "external",
                                            "profiling": "false",
                                        },
                                        "extraVolumes": [
                                            # Note(oleks): Add this as default as a workaround of the json patch limitation # noqa: E501
                                            # https://cluster-api.sigs.k8s.io/tasks/experimental-features/cluster-class/write-clusterclass#json-patches-tips--tricks
                                            {
                                                "name": "webhooks",
                                                "hostPath": "/etc/kubernetes/webhooks",
                                                "mountPath": "/etc/kubernetes/webhooks",
                                            }
                                        ],
                                    },
                                    "etcd": {
                                        "local": {
                                            "extraArgs": {
                                                "listen-metrics-urls": "http://0.0.0.0:2381",
                                            },
                                        },
                                    },
                                    "controllerManager": {
                                        "extraArgs": {
                                            "bind-address": "0.0.0.0",
                                            "cloud-provider": "external",
                                            "profiling": "false",
                                        },
                                    },
                                    "scheduler": {
                                        "extraArgs": {
                                            "bind-address": "0.0.0.0",
                                            "profiling": "false",
                                        },
                                    },
                                },
                                "files": [
                                    {
                                        "path": "/etc/kubernetes/audit-policy/apiserver-audit-policy.yaml",
                                        "permissions": "0600",
                                        "content": base64.encode_as_text(audit_policy),
                                        "encoding": "base64",
                                    },
                                    {
                                        "path": "/etc/kubernetes/webhooks/webhookconfig.yaml",
                                        "owner": "root:root",
                                        "permissions": "0644",
                                        "content": base64.encode_as_text(
                                            keystone_auth_webhook
                                        ),
                                        "encoding": "base64",
                                    },
                                    {
                                        "path": "/run/kubeadm/configure-kube-proxy.sh",
                                        "permissions": "0755",
                                        "content": base64.encode_as_text(
                                            importlib.resources.files(
                                                "magnum_cluster_api.files.run.kubeadm"
                                            )
                                            .joinpath("configure-kube-proxy.sh")
                                            .read_text()
                                        ),
                                        "encoding": "base64",
                                    },
                                ],
                                "initConfiguration": {
                                    "nodeRegistration": {
                                        "name": "{{ local_hostname }}",
                                        "kubeletExtraArgs": {
                                            "cloud-provider": "external",
                                        },
                                    },
                                },
                                "joinConfiguration": {
                                    "nodeRegistration": {
                                        "name": "{{ local_hostname }}",
                                        "kubeletExtraArgs": {
                                            "cloud-provider": "external",
                                        },
                                    },
                                },
                                "preKubeadmCommands": [
                                    "rm /var/lib/etcd/lost+found -rf",
                                    "bash /run/kubeadm/configure-kube-proxy.sh",
                                ],
                            },
                        },
                    },
                },
            },
        )


class KubeadmConfigTemplate(Base):
    def get_object(self) -> objects.KubeadmConfigTemplate:
        return objects.KubeadmConfigTemplate(
            self.api,
            {
                "apiVersion": objects.KubeadmConfigTemplate.version,
                "kind": objects.KubeadmConfigTemplate.kind,
                "metadata": {
                    "name": CLUSTER_CLASS_NAME,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "template": {
                        "spec": {
                            "files": [
                                {
                                    "path": "/etc/kubernetes/.placeholder",
                                    "permissions": "0644",
                                    "content": base64.encode_as_text(PLACEHOLDER),
                                    "encoding": "base64",
                                },
                            ],
                            "joinConfiguration": {
                                "nodeRegistration": {
                                    "name": "{{ local_hostname }}",
                                    "kubeletExtraArgs": {
                                        "cloud-provider": "external",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        )


class OpenStackMachineTemplate(Base):
    def get_object(self) -> objects.OpenStackMachineTemplate:
        return objects.OpenStackMachineTemplate(
            self.api,
            {
                "apiVersion": objects.OpenStackMachineTemplate.version,
                "kind": objects.OpenStackMachineTemplate.kind,
                "metadata": {
                    "name": CLUSTER_CLASS_NAME,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "template": {
                        "spec": {
                            "cloudName": "default",
                            "flavor": PLACEHOLDER,
                        }
                    }
                },
            },
        )


class OpenStackClusterTemplate(Base):
    def get_object(self) -> objects.OpenStackClusterTemplate:
        return objects.OpenStackClusterTemplate(
            self.api,
            {
                "apiVersion": objects.OpenStackClusterTemplate.version,
                "kind": objects.OpenStackClusterTemplate.kind,
                "metadata": {
                    "name": CLUSTER_CLASS_NAME,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "template": {
                        "spec": {
                            "cloudName": "default",
                            "managedSecurityGroups": True,
                            "allowAllInClusterTraffic": True,
                        },
                    },
                },
            },
        )


class ClusterClass(Base):
    def get_object(self) -> objects.ClusterClass:
        return objects.ClusterClass(
            self.api,
            {
                "apiVersion": objects.ClusterClass.version,
                "kind": objects.ClusterClass.kind,
                "metadata": {
                    "name": CLUSTER_CLASS_NAME,
                    "namespace": "magnum-system",
                },
                "spec": {
                    "controlPlane": {
                        "nodeVolumeDetachTimeout": CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT,
                        "ref": {
                            "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                            "kind": objects.KubeadmControlPlaneTemplate.kind,
                            "name": CLUSTER_CLASS_NAME,
                        },
                        "machineInfrastructure": {
                            "ref": {
                                "apiVersion": objects.OpenStackMachineTemplate.version,
                                "kind": objects.OpenStackMachineTemplate.kind,
                                "name": CLUSTER_CLASS_NAME,
                            },
                        },
                        "machineHealthCheck": {
                            "maxUnhealthy": "80%",
                            "unhealthyConditions": [
                                {
                                    "type": "Ready",
                                    "status": "False",
                                    "timeout": "5m",
                                },
                                {
                                    "type": "Ready",
                                    "status": "Unknown",
                                    "timeout": "5m",
                                },
                            ],
                        },
                    },
                    "infrastructure": {
                        "ref": {
                            "apiVersion": objects.OpenStackClusterTemplate.version,
                            "kind": objects.OpenStackClusterTemplate.kind,
                            "name": CLUSTER_CLASS_NAME,
                        },
                    },
                    "workers": {
                        "machineDeployments": [
                            {
                                "class": "default-worker",
                                "nodeVolumeDetachTimeout": CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT,
                                "template": {
                                    "bootstrap": {
                                        "ref": {
                                            "apiVersion": objects.KubeadmConfigTemplate.version,
                                            "kind": objects.KubeadmConfigTemplate.kind,
                                            "name": CLUSTER_CLASS_NAME,
                                        }
                                    },
                                    "infrastructure": {
                                        "ref": {
                                            "apiVersion": objects.OpenStackMachineTemplate.version,
                                            "kind": objects.OpenStackMachineTemplate.kind,
                                            "name": CLUSTER_CLASS_NAME,
                                        }
                                    },
                                },
                                "machineHealthCheck": {
                                    "maxUnhealthy": "80%",
                                    "unhealthyConditions": [
                                        {
                                            "type": "Ready",
                                            "status": "False",
                                            "timeout": "5m",
                                        },
                                        {
                                            "type": "Ready",
                                            "status": "Unknown",
                                            "timeout": "5m",
                                        },
                                    ],
                                },
                            }
                        ],
                    },
                    "variables": [
                        {
                            "name": "apiServerLoadBalancer",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "object",
                                    "required": ["enabled"],
                                    "properties": {
                                        "enabled": {
                                            "type": "boolean",
                                        },
                                        "provider": {
                                            "type": "string",
                                        },
                                    },
                                },
                            },
                        },
                        {
                            "name": "apiServerTLSCipherSuites",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "openidConnect",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "object",
                                    "required": [
                                        "issuerUrl",
                                        "clientId",
                                        "usernameClaim",
                                        "usernamePrefix",
                                        "groupsClaim",
                                        "groupsPrefix",
                                    ],
                                    "properties": {
                                        "issuerUrl": {
                                            "type": "string",
                                        },
                                        "clientId": {
                                            "type": "string",
                                        },
                                        "usernameClaim": {
                                            "type": "string",
                                        },
                                        "usernamePrefix": {
                                            "type": "string",
                                        },
                                        "groupsClaim": {
                                            "type": "string",
                                        },
                                        "groupsPrefix": {
                                            "type": "string",
                                        },
                                    },
                                },
                            },
                        },
                        {
                            "name": "auditLog",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "object",
                                    "required": [
                                        "enabled",
                                        "maxAge",
                                        "maxBackup",
                                        "maxSize",
                                    ],
                                    "properties": {
                                        "enabled": {
                                            "type": "boolean",
                                        },
                                        "maxAge": {
                                            "type": "string",
                                        },
                                        "maxBackup": {
                                            "type": "string",
                                        },
                                        "maxSize": {
                                            "type": "string",
                                        },
                                    },
                                },
                            },
                        },
                        {
                            "name": "bootVolume",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "object",
                                    "required": ["size"],
                                    "properties": {
                                        "size": {
                                            "type": "integer",
                                        },
                                        "type": {
                                            "type": "string",
                                        },
                                    },
                                },
                            },
                        },
                        {
                            "name": "clusterIdentityRef",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "object",
                                    "required": ["kind", "name"],
                                    "properties": {
                                        "kind": {
                                            "type": "string",
                                            "enum": [pykube.Secret.kind],
                                        },
                                        "name": {"type": "string"},
                                    },
                                },
                            },
                        },
                        {
                            "name": "cloudCaCert",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "cloudControllerManagerConfig",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "systemdProxyConfig",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "aptProxyConfig",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "containerdConfig",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "controlPlaneFlavor",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "disableAPIServerFloatingIP",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "boolean",
                                },
                            },
                        },
                        {
                            "name": "dnsNameservers",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "array",
                                    "items": {
                                        "type": "string",
                                    },
                                },
                            },
                        },
                        {
                            "name": "externalNetworkId",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "fixedNetworkName",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "fixedSubnetId",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "flavor",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "imageRepository",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "imageUUID",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "kubeletTLSCipherSuites",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "nodeCidr",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "sshKeyName",
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "operatingSystem",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                    "enum": utils.AVAILABLE_OPERATING_SYSTEMS,
                                    "default": "ubuntu",
                                },
                            },
                        },
                        {
                            "name": "enableDockerVolume",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "boolean",
                                },
                            },
                        },
                        {
                            "name": "dockerVolumeSize",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "integer",
                                },
                            },
                        },
                        {
                            "name": "dockerVolumeType",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "enableEtcdVolume",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "boolean",
                                },
                            },
                        },
                        {
                            "name": "etcdVolumeSize",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "integer",
                                },
                            },
                        },
                        {
                            "name": "etcdVolumeType",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "availabilityZone",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "string",
                                },
                            },
                        },
                        {
                            "name": "enableKeystoneAuth",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "boolean",
                                    "default": False,
                                },
                            },
                        },
                        {
                            "name": "controlPlaneAvailabilityZones",
                            "required": True,
                            "schema": {
                                "openAPIV3Schema": {
                                    "type": "array",
                                    "items": {
                                        "type": "string",
                                    },
                                },
                            },
                        },
                    ],
                    "patches": [
                        {
                            "name": "auditLog",
                            "enabledIf": "{{ if .auditLog.enabled }}true{{end}}",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/audit-log-path",  # noqa: E501
                                            "value": "/var/log/audit/kube-apiserver-audit.log",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/audit-log-maxage",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "auditLog.maxAge",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/audit-log-maxbackup",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "auditLog.maxBackup",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/audit-log-maxsize",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "auditLog.maxSize",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/audit-policy-file",  # noqa: E501
                                            "value": "/etc/kubernetes/audit-policy/apiserver-audit-policy.yaml",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraVolumes/-",  # noqa: E501
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    name: audit-policy
                                                    hostPath: /etc/kubernetes/audit-policy
                                                    mountPath: /etc/kubernetes/audit-policy
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraVolumes/-",  # noqa: E501
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    name: audit-logs
                                                    hostPath: /var/log/kubernetes/audit
                                                    mountPath: /var/log/audit
                                                    """
                                                ),
                                            },
                                        },
                                    ],
                                }
                            ],
                        },
                        {
                            "name": "openidConnect",
                            "enabledIf": "{{ if .openidConnect.issuerUrl }}true{{end}}",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-issuer-url",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.issuerUrl",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-client-id",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.clientId",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-username-claim",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.usernameClaim",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-username-prefix",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.usernamePrefix",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-groups-claim",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.groupsClaim",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/oidc-groups-prefix",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "openidConnect.groupsPrefix",
                                            },
                                        },
                                    ],
                                }
                            ],
                        },
                        {
                            "name": "bootVolume",
                            "enabledIf": "{{ if gt .bootVolume.size 0.0 }}true{{end}}",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackMachineTemplate.version,
                                        "kind": objects.OpenStackMachineTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            },
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/rootVolume",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    diskSize: {{ .bootVolume.size }}
                                                    volumeType: {{ .bootVolume.type }}
                                                    """
                                                ),
                                            },
                                        },
                                    ],
                                }
                            ],
                        },
                        {
                            "name": "ubuntu",
                            "enabledIf": '{{ if eq .operatingSystem "ubuntu" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/apt/apt.conf.d/90proxy"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .aptProxyConfig }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/preKubeadmCommands/-",
                                            "value": "systemctl daemon-reload && systemctl restart containerd",
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmConfigTemplate.version,
                                        "kind": objects.KubeadmConfigTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            }
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/apt/apt.conf.d/90proxy"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .aptProxyConfig }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/preKubeadmCommands",
                                            "value": [
                                                "systemctl daemon-reload",
                                                "systemctl restart containerd",
                                            ],
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "flatcar",
                            "enabledIf": '{{ if eq .operatingSystem "flatcar" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/preKubeadmCommands/-",
                                            "value": textwrap.dedent(
                                                """\
                                            bash -c "sed -i 's/__REPLACE_NODE_NAME__/$(hostname -s)/g' /etc/kubeadm.yml"
                                            bash -c "test -f /tmp/containerd-bootstrap || (touch /tmp/containerd-bootstrap && systemctl daemon-reload && systemctl restart containerd)"
                                            """  # noqa: E501
                                            ),
                                        },
                                        {
                                            "op": "replace",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/format",
                                            "value": "ignition",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/ignition",
                                            "value": {
                                                "containerLinuxConfig": {
                                                    "additionalConfig": textwrap.dedent(
                                                        """\
                                                        systemd:
                                                          units:
                                                          - name: coreos-metadata-sshkeys@.service
                                                            enabled: true
                                                          - name: kubeadm.service
                                                            enabled: true
                                                            dropins:
                                                            - name: 10-flatcar.conf
                                                              contents: |
                                                                [Unit]
                                                                Requires=containerd.service coreos-metadata.service
                                                                After=containerd.service coreos-metadata.service
                                                                [Service]
                                                                EnvironmentFile=/run/metadata/flatcar
                                                        """  # noqa: E501
                                                    ),
                                                },
                                            },
                                        },
                                        {
                                            "op": "replace",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/initConfiguration/nodeRegistration/name",  # noqa: E501
                                            "value": "__REPLACE_NODE_NAME__",
                                        },
                                        {
                                            "op": "replace",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/joinConfiguration/nodeRegistration/name",  # noqa: E501
                                            "value": "__REPLACE_NODE_NAME__",
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmConfigTemplate.version,
                                        "kind": objects.KubeadmConfigTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            }
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/preKubeadmCommands",
                                            "value": [
                                                textwrap.dedent(
                                                    """\
                                                bash -c "sed -i 's/__REPLACE_NODE_NAME__/$(hostname -s)/g' /etc/kubeadm.yml"
                                                bash -c "test -f /tmp/containerd-bootstrap || (touch /tmp/containerd-bootstrap && systemctl daemon-reload && systemctl restart containerd)"
                                                """  # noqa: E501
                                                )
                                            ],
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/format",
                                            "value": "ignition",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/ignition",
                                            "value": {
                                                "containerLinuxConfig": {
                                                    "additionalConfig": textwrap.dedent(
                                                        """\
                                                        systemd:
                                                          units:
                                                          - name: coreos-metadata-sshkeys@.service
                                                            enabled: true
                                                          - name: kubeadm.service
                                                            enabled: true
                                                            dropins:
                                                            - name: 10-flatcar.conf
                                                              contents: |
                                                                [Unit]
                                                                Requires=containerd.service coreos-metadata.service
                                                                After=containerd.service coreos-metadata.service
                                                                [Service]
                                                                EnvironmentFile=/run/metadata/flatcar
                                                        """  # noqa: E501
                                                    ),
                                                },
                                            },
                                        },
                                        {
                                            "op": "replace",
                                            "path": "/spec/template/spec/joinConfiguration/nodeRegistration/name",
                                            "value": "__REPLACE_NODE_NAME__",
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "clusterConfig",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackMachineTemplate.version,
                                        "kind": objects.OpenStackMachineTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/flavor",
                                            "valueFrom": {
                                                "variable": "controlPlaneFlavor",
                                            },
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackMachineTemplate.version,
                                        "kind": objects.OpenStackMachineTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            },
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/flavor",
                                            "valueFrom": {
                                                "variable": "flavor",
                                            },
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackMachineTemplate.version,
                                        "kind": objects.OpenStackMachineTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            },
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/identityRef",
                                            "valueFrom": {
                                                "variable": "clusterIdentityRef"
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/sshKeyName",
                                            "valueFrom": {"variable": "sshKeyName"},
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/imageUUID",
                                            "valueFrom": {"variable": "imageUUID"},
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackClusterTemplate.version,
                                        "kind": objects.OpenStackClusterTemplate.kind,
                                        "matchResources": {
                                            "infrastructureCluster": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/apiServerLoadBalancer",
                                            "valueFrom": {
                                                "variable": "apiServerLoadBalancer"
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/identityRef",
                                            "valueFrom": {
                                                "variable": "clusterIdentityRef"
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/disableAPIServerFloatingIP",
                                            "valueFrom": {
                                                "variable": "disableAPIServerFloatingIP"
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/dnsNameservers",
                                            "valueFrom": {"variable": "dnsNameservers"},
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/externalNetworkId",
                                            "valueFrom": {
                                                "variable": "externalNetworkId"
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "controlPlaneAvailabilityZones",
                            "enabledIf": '{{ if ne (index .controlPlaneAvailabilityZones 0) "" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackClusterTemplate.version,
                                        "kind": objects.OpenStackClusterTemplate.kind,
                                        "matchResources": {
                                            "infrastructureCluster": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/controlPlaneAvailabilityZones",
                                            "valueFrom": {
                                                "variable": "controlPlaneAvailabilityZones"
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "newNetworkConfig",
                            "enabledIf": '{{ if eq .fixedNetworkName "" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackClusterTemplate.version,
                                        "kind": objects.OpenStackClusterTemplate.kind,
                                        "matchResources": {
                                            "infrastructureCluster": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/nodeCidr",
                                            "valueFrom": {"variable": "nodeCidr"},
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "existingFixedNetworkNameConfig",
                            "enabledIf": '{{ if ne .fixedNetworkName "" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackClusterTemplate.version,
                                        "kind": objects.OpenStackClusterTemplate.kind,
                                        "matchResources": {
                                            "infrastructureCluster": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/network/name",
                                            "valueFrom": {
                                                "variable": "fixedNetworkName"
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "existingFixedSubnetIdConfig",
                            "enabledIf": '{{ if ne .fixedSubnetId "" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.OpenStackClusterTemplate.version,
                                        "kind": objects.OpenStackClusterTemplate.kind,
                                        "matchResources": {
                                            "infrastructureCluster": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/subnet/id",
                                            "valueFrom": {"variable": "fixedSubnetId"},
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "customImageRepository",
                            "enabledIf": '{{ if ne .imageRepository "" }}true{{end}}',
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/imageRepository",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "imageRepository",
                                            },
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmConfigTemplate.version,
                                        "kind": objects.KubeadmConfigTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            }
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/clusterConfiguration",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    imageRepository: "{{ .imageRepository }}"
                                                    """
                                                ),
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "etcdVolumeAndDockerVolume",
                            "enabledIf": "{{ if and .enableEtcdVolume .enableDockerVolume }}true{{ end }}",
                            "definitions": json_patches.Volumes(
                                control_plane_disks=[
                                    json_patches.DiskConfig(
                                        type="etcd",
                                        mount_path="/var/lib/etcd",
                                    ),
                                    json_patches.DiskConfig(
                                        type="docker",
                                        mount_path="/var/lib/containerd",
                                    ),
                                ],
                                worker_disks=[
                                    json_patches.DiskConfig(
                                        type="docker",
                                        mount_path="/var/lib/containerd",
                                    )
                                ],
                            ).definitions,
                        },
                        {
                            "name": "onlyDockerVolume",
                            "enabledIf": "{{ if and .enableDockerVolume (not .enableEtcdVolume) }}true{{ end }}",
                            "definitions": json_patches.Volumes(
                                control_plane_disks=[
                                    json_patches.DiskConfig(
                                        type="docker",
                                        mount_path="/var/lib/containerd",
                                    )
                                ],
                                worker_disks=[
                                    json_patches.DiskConfig(
                                        type="docker",
                                        mount_path="/var/lib/containerd",
                                    )
                                ],
                            ).definitions,
                        },
                        {
                            "name": "onlyEtcdVolume",
                            "enabledIf": "{{ if and .enableEtcdVolume (not .enableDockerVolume) }}true{{ end }}",
                            "definitions": json_patches.Volumes(
                                control_plane_disks=[
                                    json_patches.DiskConfig(
                                        type="etcd",
                                        mount_path="/var/lib/etcd",
                                    )
                                ],
                            ).definitions,
                        },
                        {
                            "name": "keystoneAuth",
                            "enabledIf": "{{ if .enableKeystoneAuth }}true{{end}}",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/authentication-token-webhook-config-file",  # noqa: E501
                                            "value": "/etc/kubernetes/webhooks/webhookconfig.yaml",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/authorization-webhook-config-file",  # noqa: E501
                                            "value": "/etc/kubernetes/webhooks/webhookconfig.yaml",
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/authorization-mode",  # noqa: E501
                                            "value": "Node,RBAC,Webhook",
                                        },
                                    ],
                                }
                            ],
                        },
                        {
                            "name": "controlPlaneConfig",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmControlPlaneTemplate.version,
                                        "kind": objects.KubeadmControlPlaneTemplate.kind,
                                        "matchResources": {
                                            "controlPlane": True,
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/extraArgs/tls-cipher-suites",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "apiServerTLSCipherSuites",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/initConfiguration/nodeRegistration/kubeletExtraArgs/tls-cipher-suites",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "kubeletTLSCipherSuites",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/joinConfiguration/nodeRegistration/kubeletExtraArgs/tls-cipher-suites",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "kubeletTLSCipherSuites",
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/clusterConfiguration/apiServer/certSANs",  # noqa: E501
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    - {{ .builtin.cluster.name }}
                                                    - {{ .builtin.cluster.name }}.{{ .builtin.cluster.namespace }}
                                                    - {{ .builtin.cluster.name }}.{{ .builtin.cluster.namespace }}.svc
                                                    - {{ .builtin.cluster.name }}.{{ .builtin.cluster.namespace }}.svc.cluster.local # noqa: E501
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/containerd/config.toml"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .containerdConfig }}"
                                                    encoding: "base64"
                                                    """
                                                )
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/kubernetes/cloud.conf"
                                                    owner: "root:root"
                                                    permissions: "0600"
                                                    content: "{{ .cloudControllerManagerConfig }}"
                                                    encoding: "base64"
                                                    """
                                                )
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/kubernetes/cloud_ca.crt"
                                                    owner: "root:root"
                                                    permissions: "0600"
                                                    content: "{{ .cloudCaCert }}"
                                                    encoding: "base64"
                                                    """
                                                )
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/kubeadmConfigSpec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/systemd/system/containerd.service.d/proxy.conf"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .systemdProxyConfig }}"
                                                    encoding: "base64"
                                                    """
                                                )
                                            },
                                        },
                                    ],
                                },
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmConfigTemplate.version,
                                        "kind": objects.KubeadmConfigTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            }
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/kubernetes/cloud.conf"
                                                    owner: "root:root"
                                                    permissions: "0600"
                                                    content: "{{ .cloudControllerManagerConfig }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/kubernetes/cloud_ca.crt"
                                                    owner: "root:root"
                                                    permissions: "0600"
                                                    content: "{{ .cloudCaCert }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/containerd/config.toml"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .containerdConfig }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/files/-",
                                            "valueFrom": {
                                                "template": textwrap.dedent(
                                                    """\
                                                    path: "/etc/systemd/system/containerd.service.d/proxy.conf"
                                                    owner: "root:root"
                                                    permissions: "0644"
                                                    content: "{{ .systemdProxyConfig }}"
                                                    encoding: "base64"
                                                    """
                                                ),
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                        {
                            "name": "workerConfig",
                            "definitions": [
                                {
                                    "selector": {
                                        "apiVersion": objects.KubeadmConfigTemplate.version,
                                        "kind": objects.KubeadmConfigTemplate.kind,
                                        "matchResources": {
                                            "machineDeploymentClass": {
                                                "names": ["default-worker"],
                                            }
                                        },
                                    },
                                    "jsonPatches": [
                                        {
                                            "op": "add",
                                            "path": "/spec/template/spec/joinConfiguration/nodeRegistration/kubeletExtraArgs/tls-cipher-suites",  # noqa: E501
                                            "valueFrom": {
                                                "variable": "kubeletTLSCipherSuites",
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                    ],
                },
            },
        )


def create_cluster_class(
    api: pykube.HTTPClient,
) -> ClusterClass:
    """
    Create a ClusterClass and all of it's supporting resources from a Magnum
    cluster template using server-side apply.
    """

    KubeadmControlPlaneTemplate(api).apply()
    KubeadmConfigTemplate(api).apply()
    OpenStackMachineTemplate(api).apply()
    OpenStackClusterTemplate(api).apply()
    ClusterClass(api).apply()


def mutate_machine_deployment(
    context: context.RequestContext,
    cluster: objects.Cluster,
    node_group: magnum_objects.NodeGroup,
    machine_deployment: dict = None,
):
    """
    This function will either makes updates to machine deployment fields which
    will not cause a rolling update or will return a new machine deployment
    if none is provided.
    """

    # NOTE(okozachenko1203): Initialize as an empty dict if not provided
    #                        instead of using mutable default argument.
    if machine_deployment is None:
        machine_deployment = {}

    auto_scaling_enabled = utils.get_auto_scaling_enabled(cluster)

    machine_deployment.setdefault(
        "metadata",
        {
            "annotations": {},
            "labels": {},
        },
    )

    # Node labels
    machine_deployment["metadata"]["labels"] = {
        f"node-role.kubernetes.io/{node_group.role}": "",
        "node.cluster.x-k8s.io/nodegroup": node_group.name,
    }

    # Replicas (or min/max if auto-scaling is enabled)
    if auto_scaling_enabled:
        machine_deployment["replicas"] = None
        machine_deployment["metadata"]["annotations"] = {
            AUTOSCALE_ANNOTATION_MIN: str(
                utils.get_node_group_min_node_count(node_group)
            ),
            AUTOSCALE_ANNOTATION_MAX: str(
                utils.get_node_group_max_node_count(node_group)
            ),
        }
    else:
        machine_deployment["replicas"] = node_group.node_count
        machine_deployment["metadata"]["annotations"] = {}

    # Fixes
    machine_deployment["nodeVolumeDetachTimeout"] = (
        CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT
    )

    # Anything beyond this point will *NOT* be changed in the machine deployment
    # for update operations (i.e. if the machine deployment already exists).
    if machine_deployment.get("name") == node_group.name:
        return machine_deployment

    # At this point, this is all code that will be added for brand new machine
    # deployments.  We can bring any of this code into the above block if we
    # want to change it for existing machine deployments.
    machine_deployment.update(
        {
            "class": "default-worker",
            "name": node_group.name,
            "failureDomain": node_group.labels.get("availability_zone", ""),
            "machineHealthCheck": {"enable": utils.get_auto_healing_enabled(cluster)},
            "variables": {
                "overrides": [
                    {
                        "name": "bootVolume",
                        "value": {
                            "size": utils.get_node_group_label_as_int(
                                node_group,
                                "boot_volume_size",
                                CONF.cinder.default_boot_volume_size,
                            ),
                            "type": node_group.labels.get(
                                "boot_volume_type",
                                cinder.get_default_boot_volume_type(context),
                            ),
                        },
                    },
                    {
                        "name": "flavor",
                        "value": node_group.flavor_id,
                    },
                    {
                        "name": "imageRepository",
                        "value": node_group.labels.get(
                            "container_infra_prefix",
                            "",
                        ),
                    },
                    {
                        "name": "imageUUID",
                        "value": utils.get_image_uuid(node_group.image_id, context),
                    },
                ],
            },
        }
    )

    return machine_deployment


def generate_machine_deployments_for_cluster(
    context: context.RequestContext, cluster: objects.Cluster
) -> list:
    machine_deployments = []
    for ng in cluster.nodegroups:
        if ng.role == "master" or ng.status.startswith("DELETE"):
            continue

        machine_deployment = mutate_machine_deployment(context, cluster, ng)
        machine_deployments.append(machine_deployment)

    return machine_deployments


class Cluster(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
    ):
        self.context = context
        self.api = api
        self.cluster = cluster

    @property
    def labels(self) -> dict:
        cni_version = self.cluster.labels.get("calico_tag", CALICO_TAG)
        labels = {
            "cni": f"calico-{cni_version}",
        }

        return {**super().labels, **labels}

    def get_or_none(self) -> objects.Cluster:
        return objects.Cluster.objects(self.api, namespace="magnum-system").get_or_none(
            name=self.cluster.stack_id
        )

    def get_object(self) -> objects.Cluster:
        osc = clients.get_openstack_api(self.context)
        default_volume_type = osc.cinder().volume_types.default()
        return objects.Cluster(
            self.api,
            {
                "apiVersion": objects.Cluster.version,
                "kind": objects.Cluster.kind,
                "metadata": {
                    "name": self.cluster.stack_id,
                    "namespace": "magnum-system",
                    "labels": self.labels,
                },
                "spec": {
                    "clusterNetwork": {
                        "serviceDomain": self.cluster.labels.get(
                            "dns_cluster_domain", "cluster.local"
                        ),
                        "pods": {
                            "cidrBlocks": [
                                self.cluster.labels.get(
                                    "calico_ipv4pool", "10.100.0.0/16"
                                )
                            ],
                        },
                        "services": {
                            "cidrBlocks": [
                                self.cluster.labels.get(
                                    "service_cluster_ip_range", "10.254.0.0/16"
                                )
                            ],
                        },
                    },
                    "topology": {
                        "class": CLUSTER_CLASS_NAME,
                        "version": utils.get_kube_tag(self.cluster),
                        "controlPlane": {
                            "metadata": {
                                "labels": {
                                    "node-role.kubernetes.io/master": "",
                                }
                            },
                            "replicas": self.cluster.master_count,
                            "machineHealthCheck": {
                                "enable": utils.get_auto_healing_enabled(self.cluster)
                            },
                        },
                        "workers": {
                            "machineDeployments": generate_machine_deployments_for_cluster(
                                self.context, self.cluster
                            ),
                        },
                        "variables": [
                            {
                                "name": "apiServerLoadBalancer",
                                "value": {
                                    "enabled": self.cluster.master_lb_enabled,
                                    "provider": self.cluster.labels.get(
                                        "octavia_provider", "amphora"
                                    ),
                                },
                            },
                            {
                                "name": "apiServerTLSCipherSuites",
                                "value": self.cluster.labels.get(
                                    "api_server_tls_cipher_suites",
                                    "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305,TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305",  # noqa: E501
                                ),
                            },
                            {
                                "name": "openidConnect",
                                "value": {
                                    "clientId": self.cluster.labels.get(
                                        "oidc_client_id", ""
                                    ),
                                    "groupsClaim": self.cluster.labels.get(
                                        "oidc_groups_claim", ""
                                    ),
                                    "groupsPrefix": self.cluster.labels.get(
                                        "oidc_groups_prefix", ""
                                    ),
                                    "issuerUrl": self.cluster.labels.get(
                                        "oidc_issuer_url", ""
                                    ),
                                    "usernameClaim": self.cluster.labels.get(
                                        "oidc_username_claim", "sub"
                                    ),
                                    "usernamePrefix": self.cluster.labels.get(
                                        "oidc_username_prefix", "-"
                                    ),
                                },
                            },
                            {
                                "name": "auditLog",
                                "value": {
                                    "enabled": utils.get_cluster_label_as_bool(
                                        self.cluster, "audit_log_enabled", False
                                    ),
                                    "maxAge": self.cluster.labels.get(
                                        "audit_log_max_age", "30"
                                    ),
                                    "maxBackup": self.cluster.labels.get(
                                        "audit_log_max_backup", "10"
                                    ),
                                    "maxSize": self.cluster.labels.get(
                                        "audit_log_max_size", "100"
                                    ),
                                },
                            },
                            {
                                "name": "bootVolume",
                                "value": {
                                    "size": utils.get_cluster_label_as_int(
                                        self.cluster,
                                        "boot_volume_size",
                                        CONF.cinder.default_boot_volume_size,
                                    ),
                                    "type": self.cluster.labels.get(
                                        "boot_volume_type",
                                        cinder.get_default_boot_volume_type(
                                            self.context
                                        ),
                                    ),
                                },
                            },
                            {
                                "name": "clusterIdentityRef",
                                "value": {
                                    "kind": pykube.Secret.kind,
                                    "name": utils.get_cluster_api_cloud_config_secret_name(
                                        self.cluster
                                    ),
                                },
                            },
                            {
                                "name": "cloudCaCert",
                                "value": base64.encode_as_text(
                                    utils.get_cloud_ca_cert()
                                ),
                            },
                            {
                                "name": "cloudControllerManagerConfig",
                                "value": base64.encode_as_text(
                                    utils.generate_cloud_controller_manager_config(
                                        self.context, self.api, self.cluster
                                    )
                                ),
                            },
                            {
                                "name": "systemdProxyConfig",
                                "value": base64.encode_as_text(
                                    utils.generate_systemd_proxy_config(self.cluster)
                                ),
                            },
                            {
                                "name": "aptProxyConfig",
                                "value": base64.encode_as_text(
                                    utils.generate_apt_proxy_config(self.cluster)
                                ),
                            },
                            {
                                "name": "containerdConfig",
                                "value": base64.encode_as_text(
                                    utils.generate_containerd_config(self.cluster)
                                ),
                            },
                            {
                                "name": "controlPlaneFlavor",
                                "value": self.cluster.master_flavor_id,
                            },
                            {
                                "name": "disableAPIServerFloatingIP",
                                "value": utils.get_cluster_floating_ip_disabled(
                                    self.cluster
                                ),
                            },
                            {
                                "name": "dnsNameservers",
                                "value": self.cluster.cluster_template.dns_nameserver.split(
                                    ","
                                ),
                            },
                            {
                                "name": "externalNetworkId",
                                "value": neutron.get_external_network_id(
                                    self.context,
                                    self.cluster.cluster_template.external_network_id,
                                ),
                            },
                            {
                                "name": "fixedNetworkName",
                                "value": neutron.get_fixed_network_name(
                                    self.context, self.cluster.fixed_network
                                )
                                or "",
                            },
                            {
                                "name": "fixedSubnetId",
                                "value": neutron.get_fixed_subnet_id(
                                    self.context, self.cluster.fixed_subnet
                                )
                                or "",
                            },
                            {
                                "name": "flavor",
                                "value": self.cluster.flavor_id,
                            },
                            {
                                "name": "imageRepository",
                                "value": utils.get_cluster_container_infra_prefix(
                                    self.cluster,
                                ),
                            },
                            {
                                "name": "imageUUID",
                                "value": utils.get_image_uuid(
                                    self.cluster.default_ng_master.image_id,
                                    self.context,
                                ),
                            },
                            {
                                "name": "kubeletTLSCipherSuites",
                                "value": self.cluster.labels.get(
                                    "kubelet_tls_cipher_suites",
                                    "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305",  # noqa: E501
                                ),
                            },
                            {
                                "name": "nodeCidr",
                                "value": self.cluster.labels.get(
                                    "fixed_subnet_cidr",
                                    "10.0.0.0/24",
                                ),
                            },
                            {
                                "name": "sshKeyName",
                                "value": self.cluster.keypair or "",
                            },
                            {
                                "name": "operatingSystem",
                                "value": utils.get_operating_system(self.cluster),
                            },
                            {
                                "name": "enableDockerVolume",
                                "value": utils.get_cluster_label_as_int(
                                    self.cluster,
                                    "docker_volume_size",
                                    0,
                                )
                                > 0,
                            },
                            {
                                "name": "dockerVolumeSize",
                                "value": utils.get_cluster_label_as_int(
                                    self.cluster,
                                    "docker_volume_size",
                                    0,
                                ),
                            },
                            {
                                "name": "dockerVolumeType",
                                "value": self.cluster.labels.get(
                                    "docker_volume_type",
                                    default_volume_type.name,
                                ),
                            },
                            {
                                "name": "enableEtcdVolume",
                                "value": utils.get_cluster_label_as_int(
                                    self.cluster,
                                    "etcd_volume_size",
                                    0,
                                )
                                > 0,
                            },
                            {
                                "name": "etcdVolumeSize",
                                "value": utils.get_cluster_label_as_int(
                                    self.cluster,
                                    "etcd_volume_size",
                                    0,
                                ),
                            },
                            {
                                "name": "etcdVolumeType",
                                "value": self.cluster.labels.get(
                                    "etcd_volume_type",
                                    default_volume_type.name,
                                ),
                            },
                            {
                                "name": "availabilityZone",
                                "value": self.cluster.labels.get(
                                    "availability_zone", ""
                                ),
                            },
                            {
                                "name": "enableKeystoneAuth",
                                "value": utils.get_cluster_label_as_bool(
                                    self.cluster, "keystone_auth_enabled", True
                                ),
                            },
                            {
                                "name": "controlPlaneAvailabilityZones",
                                "value": self.cluster.labels.get(
                                    "control_plane_availability_zones", ""
                                ).split(","),
                            },
                        ],
                    },
                },
            },
        )

    def delete(self):
        capi_cluster = self.get_or_none()
        if capi_cluster:
            capi_cluster.delete()


def apply_cluster_from_magnum_cluster(
    context: context.RequestContext,
    api: pykube.HTTPClient,
    cluster: magnum_objects.Cluster,
    skip_auto_scaling_release: bool = False,
) -> None:
    """
    Create a ClusterAPI cluster given a Magnum Cluster object.
    """
    create_cluster_class(api)

    ClusterResourcesConfigMap(context, api, cluster).apply()
    ClusterResourceSet(api, cluster).apply()
    Cluster(context, api, cluster).apply()

    if not skip_auto_scaling_release and utils.get_auto_scaling_enabled(cluster):
        ClusterAutoscalerHelmRelease(api, cluster).apply()


def get_kubeadm_control_plane(
    api: pykube.HTTPClient, cluster: magnum_objects.Cluster
) -> objects.KubeadmControlPlane:
    kcps = objects.KubeadmControlPlane.objects(api, namespace="magnum-system").filter(
        selector={
            "cluster.x-k8s.io/cluster-name": cluster.stack_id,
        },
    )
    if len(kcps) == 1:
        return list(kcps)[0]
    return None


def get_machine_deployment(
    api: pykube.HTTPClient,
    cluster: magnum_objects.Cluster,
    node_group: magnum_objects.NodeGroup,
) -> objects.KubeadmControlPlane:
    mds = objects.MachineDeployment.objects(api, namespace="magnum-system").filter(
        selector={
            "cluster.x-k8s.io/cluster-name": cluster.stack_id,
            "topology.cluster.x-k8s.io/deployment-name": node_group.name,
        },
    )
    if len(mds) == 1:
        return list(mds)[0]
    return None



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sherlock


class ClusterLock(sherlock.KubernetesLock):
    """
    A cluster lock that is used to lock the cluster for any operations
    across all of the conductor nodes.
    """

    DEFAULT_EXPIRE: int = 60

    def __init__(self, cluster_id: str, expire: int = DEFAULT_EXPIRE):
        sherlock.configure(
            backend=sherlock.backends.KUBERNETES,
            retry_interval=1,
        )

        super().__init__(
            lock_name="cluster-%s" % cluster_id,
            k8s_namespace="magnum-system",
            expire=expire,
        )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.conductor import monitors
from magnum.objects import fields
from oslo_log import log as logging
from oslo_utils import strutils

from magnum_cluster_api import clients, objects, utils

LOG = logging.getLogger(__name__)


class Monitor(monitors.MonitorBase):
    def metrics_spec(self):
        pass

    def pull_data(self):
        pass

    def poll_nodegroup_replicas(self):
        """
        Poll the number of replicas of each nodegroup in the cluster when autoscaling enabled.
        """
        k8s_api = clients.get_pykube_api()
        if not utils.get_auto_scaling_enabled(self.cluster):
            return
        for node_group in self.cluster.nodegroups:
            md = objects.MachineDeployment.for_node_group_or_none(
                k8s_api, self.cluster, node_group
            )
            if md is None:
                continue
            node_group.node_count = md.obj["spec"]["replicas"]
            node_group.save()

    def poll_health_status(self):
        k8s_api = clients.get_pykube_api()
        self.data = {
            "health_status": fields.ClusterHealthStatus.UNKNOWN,
            "health_status_reason": {},
        }

        machines = objects.Machine.objects(k8s_api).filter(
            namespace="magnum-system",
            selector={
                "cluster.x-k8s.io/cluster-name": self.cluster.stack_id,
            },
        )

        if len(machines) == 0:
            return

        for machine in machines:
            condition_map = {
                c["type"]: c["status"] for c in machine.obj["status"]["conditions"]
            }

            node_healthy = condition_map.get("NodeHealthy", False)
            health_status = strutils.bool_from_string(node_healthy)
            self.data["health_status_reason"][f"{machine.name}.Ready"] = health_status

            if health_status is False:
                self.data["health_status"] = fields.ClusterHealthStatus.UNHEALTHY

        if self.data["health_status"] == fields.ClusterHealthStatus.UNKNOWN:
            self.data["health_status"] = fields.ClusterHealthStatus.HEALTHY

        self.poll_nodegroup_replicas()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap

import pykube  # type: ignore
import pytest
import responses


@pytest.fixture(scope="session")
def kubeconfig(tmp_path_factory):
    kubeconfig = tmp_path_factory.mktemp("pykube") / "kubeconfig"
    kubeconfig.write_text(
        textwrap.dedent(
            """
            # TODO: Replace with a more realistic example
            current-context: thecluster
            clusters:
                - name: thecluster
                  cluster: {}
            users:
                - name: admin
                  user:
                    username: adm
                    password: somepassword
            contexts:
                - name: thecluster
                  context:
                    cluster: thecluster
                    user: admin
                - name: second
                  context: secondcontext
            """
        )
    )

    return pykube.KubeConfig.from_file(str(kubeconfig))


@pytest.fixture(scope="session")
def pykube_api(kubeconfig):
    return pykube.HTTPClient(kubeconfig)


@pytest.fixture(scope="session")
def requests_mock(session_mocker, kubeconfig):
    session_mocker.patch(
        "pykube.KubeConfig.from_env",
        return_value=kubeconfig,
    )

    return responses.RequestsMock(
        target="pykube.http.KubernetesHTTPAdapter._do_send",
    )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import pytest
from oslo_config import cfg

from magnum_cluster_api import images


@pytest.mark.parametrize(
    "version,image,image_repository",
    [
        ("v1.22.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.22.3", None),
        ("v1.22.17", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.22.3", None),
        ("v1.23.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.23.1", None),
        ("v1.23.17", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.23.1", None),
        ("v1.24.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.24.2", None),
        ("v1.24.17", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.24.2", None),
        ("v1.25.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.25.2", None),
        ("v1.25.17", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.25.2", None),
        ("v1.26.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.26.3", None),
        ("v1.26.3", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.26.3", None),
        ("v1.27.0", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.27.2", None),
        ("v1.27.3", "registry.k8s.io/autoscaling/cluster-autoscaler:v1.27.2", None),
        ("v1.22.0", "quay.io/vexxhost/cluster-autoscaler:v1.22.3", "quay.io/vexxhost"),
        ("v1.22.17", "quay.io/vexxhost/cluster-autoscaler:v1.22.3", "quay.io/vexxhost"),
        ("v1.23.0", "quay.io/vexxhost/cluster-autoscaler:v1.23.1", "quay.io/vexxhost"),
        ("v1.23.17", "quay.io/vexxhost/cluster-autoscaler:v1.23.1", "quay.io/vexxhost"),
        ("v1.24.0", "quay.io/vexxhost/cluster-autoscaler:v1.24.2", "quay.io/vexxhost"),
        ("v1.24.17", "quay.io/vexxhost/cluster-autoscaler:v1.24.2", "quay.io/vexxhost"),
        ("v1.25.0", "quay.io/vexxhost/cluster-autoscaler:v1.25.2", "quay.io/vexxhost"),
        ("v1.25.17", "quay.io/vexxhost/cluster-autoscaler:v1.25.2", "quay.io/vexxhost"),
        ("v1.26.0", "quay.io/vexxhost/cluster-autoscaler:v1.26.3", "quay.io/vexxhost"),
        ("v1.26.3", "quay.io/vexxhost/cluster-autoscaler:v1.26.3", "quay.io/vexxhost"),
        ("v1.27.0", "quay.io/vexxhost/cluster-autoscaler:v1.27.2", "quay.io/vexxhost"),
        ("v1.27.3", "quay.io/vexxhost/cluster-autoscaler:v1.27.2", "quay.io/vexxhost"),
    ],
)
def test_get_cluster_autoscaler_image(image_repository, version, image):
    if image_repository:
        cfg.CONF.set_override(
            "image_repository", image_repository, group="auto_scaling"
        )
    assert images.get_cluster_autoscaler_image(version) == image



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import textwrap

import pykube
import pytest
import responses
from magnum.tests.unit.objects import utils as magnum_test_utils  # type: ignore
from oslo_serialization import base64, jsonutils

from magnum_cluster_api import exceptions, utils


def test_generate_cluster_api_name(mocker):
    mock_cluster_exists = mocker.patch("magnum_cluster_api.utils.cluster_exists")
    mock_cluster_exists.return_value = False

    api = mocker.Mock()

    cluster_api_name = utils.generate_cluster_api_name(api)

    # NOTE(mnaser): We need to make sure that the cluster_api_name is shorter
    #               than X characters so the node names are under 63 characters
    potential_node_name = "-".join(
        [cluster_api_name, "default-worker", "abcde", "abcdefghij"]
    )

    assert len(potential_node_name) <= 63


class TestGenerateCloudControllerManagerConfig:
    @pytest.fixture(autouse=True)
    def setup(self, context, pykube_api, mocker):
        self.context = context
        self.pykube_api = pykube_api

        self.cluster = magnum_test_utils.get_test_cluster(context, labels={})
        self.cluster.cluster_template = magnum_test_utils.get_test_cluster_template(
            self.context
        )

        mock_get_openstack_api = mocker.patch(
            "magnum_cluster_api.clients.get_openstack_api"
        ).return_value
        mock_get_openstack_api.url_for.return_value = "http://localhost/v3"

    def _response_for_cloud_config_secret(self):
        return responses.Response(
            responses.GET,
            "http://localhost/api/%s/namespaces/%s/%s/%s"
            % (
                pykube.Secret.version,
                "magnum-system",
                pykube.Secret.endpoint,
                utils.get_cluster_api_cloud_config_secret_name(self.cluster),
            ),
            json={
                "data": {
                    "clouds.yaml": base64.encode_as_text(
                        jsonutils.dumps(
                            {
                                "clouds": {
                                    "default": {
                                        "region_name": "RegionOne",
                                        "auth": {
                                            "application_credential_id": "fake_application_credential_id",
                                            "application_credential_secret": "fake_application_credential_secret",
                                        },
                                    }
                                }
                            }
                        )
                    ),
                }
            },
        )

    def test_generate_cloud_controller_manager_config(self, mocker, requests_mock):
        with requests_mock as rsps:
            rsps.add(self._response_for_cloud_config_secret())

            config = utils.generate_cloud_controller_manager_config(
                self.context, self.pykube_api, self.cluster
            )

        assert config == textwrap.dedent(
            """\
            [Global]
            auth-url=http://localhost/v3
            region=RegionOne
            application-credential-id=fake_application_credential_id
            application-credential-secret=fake_application_credential_secret
            tls-insecure=false

            [LoadBalancer]
            lb-provider=amphora
            lb-method=ROUND_ROBIN
            """
        )

    def test_generate_cloud_controller_manager_config_for_amphora(self, requests_mock):
        self.cluster.labels = {"octavia_provider": "amphora"}

        with requests_mock as rsps:
            rsps.add(self._response_for_cloud_config_secret())

            config = utils.generate_cloud_controller_manager_config(
                self.context, self.pykube_api, self.cluster
            )

        assert config == textwrap.dedent(
            """\
            [Global]
            auth-url=http://localhost/v3
            region=RegionOne
            application-credential-id=fake_application_credential_id
            application-credential-secret=fake_application_credential_secret
            tls-insecure=false

            [LoadBalancer]
            lb-provider=amphora
            lb-method=ROUND_ROBIN
            """
        )

    def test_generate_cloud_controller_manager_config_for_ovn(self, requests_mock):
        self.cluster.labels = {"octavia_provider": "ovn"}

        with requests_mock as rsps:
            rsps.add(self._response_for_cloud_config_secret())

            config = utils.generate_cloud_controller_manager_config(
                self.context, self.pykube_api, self.cluster
            )

        assert config == textwrap.dedent(
            """\
            [Global]
            auth-url=http://localhost/v3
            region=RegionOne
            application-credential-id=fake_application_credential_id
            application-credential-secret=fake_application_credential_secret
            tls-insecure=false

            [LoadBalancer]
            lb-provider=ovn
            lb-method=SOURCE_IP_PORT
            """
        )

    def test_generate_cloud_controller_manager_config_for_ovn_with_correct_algorithm(
        self, requests_mock
    ):
        self.cluster.labels = {
            "octavia_provider": "ovn",
            "octavia_lb_algorithm": "SOURCE_IP_PORT",
        }

        with requests_mock as rsps:
            rsps.add(self._response_for_cloud_config_secret())

            config = utils.generate_cloud_controller_manager_config(
                self.context, self.pykube_api, self.cluster
            )

        assert config == textwrap.dedent(
            """\
            [Global]
            auth-url=http://localhost/v3
            region=RegionOne
            application-credential-id=fake_application_credential_id
            application-credential-secret=fake_application_credential_secret
            tls-insecure=false

            [LoadBalancer]
            lb-provider=ovn
            lb-method=SOURCE_IP_PORT
            """
        )

    def test_generate_cloud_controller_manager_config_for_ovn_with_invalid_algorithm(
        self, requests_mock
    ):
        self.cluster.labels = {
            "octavia_provider": "ovn",
            "octavia_lb_algorithm": "ROUND_ROBIN",
        }

        with requests_mock as rsps:
            rsps.add(self._response_for_cloud_config_secret())

            with pytest.raises(exceptions.InvalidOctaviaLoadBalancerAlgorithm):
                utils.generate_cloud_controller_manager_config(
                    self.context, self.pykube_api, self.cluster
                )



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE(mnaser): We use this file to gate if we do build new images or not,
#               so we will simply keep increasing the following placeholder,
#               feel free to keep counting up.
#
#               Image build count: 1

import getpass
import json
import os
import subprocess
import tarfile
import tempfile
import textwrap
import zlib
from os.path import expanduser
from pathlib import Path

import click
import requests

QEMU_PACKAGES = [
    "qemu-kvm",
    "qemu-utils",
]


def validate_version(_, __, value):
    if not value.startswith("v"):
        raise click.BadParameter("Version should start with 'v'")
    return value


@click.command()
@click.pass_context
@click.option(
    "--operating-system",
    show_default=True,
    default="ubuntu-2204",
    type=click.Choice(
        ["ubuntu-2004", "ubuntu-2204", "flatcar", "rockylinux-8", "rockylinux-9"]
    ),
    help="Operating system to build image for",
    prompt="Operating system to build image for",
)
@click.option(
    "--version",
    show_default=True,
    default="v1.27.8",
    callback=validate_version,
    help="Kubernetes version",
)
@click.option(
    "--image-builder-version",
    show_default=True,
    default="v0.1.29",
    help="Image builder tag (or commit) to use for building image",
)
@click.option(
    "--extra-ansible-user-vars",
    show_default=True,
    default="",
    help="Extra user defined variables to set in image-builder ansible_user_vars",
)
@click.option(
    "--node-custom-roles-pre",
    show_default=True,
    default="",
    help="Custom pre-roles to run in image-builder",
)
def main(
    ctx: click.Context,
    operating_system,
    version,
    image_builder_version,
    extra_ansible_user_vars,
    node_custom_roles_pre,
):
    ib_path = f"/tmp/image-builder-{image_builder_version}"
    output_path = f"{ib_path}/images/capi/output"

    # Scan the entire output directory recursively and stop if there is any file
    # at all
    for root, dirs, files in os.walk(output_path):
        if files:
            message = (
                "There are files in the output directory which will cause the build to fail. "
                "Please remove them before continuing.\n"
            )
            for file in files:
                message += f"- {root}/{file}\n"

            ctx.fail(message)

    click.echo("- Update apt")
    subprocess.run(
        ["sudo", "/usr/bin/apt-get", "update", "-y"],
        check=True,
    )

    click.echo("- Install QEMU packages")
    subprocess.run(
        ["sudo", "/usr/bin/apt", "install", "--no-install-recommends", "-y"]
        + QEMU_PACKAGES,
        check=True,
    )

    click.echo("- Add current user to KVM group")
    subprocess.run(
        ["sudo", "/usr/sbin/usermod", "-a", "-G", "kvm", getpass.getuser()], check=True
    )

    click.echo("- Update permissions for the KVM device")
    subprocess.run(["sudo", "/bin/chown", "root:kvm", "/dev/kvm"], check=True)

    # Setup our stream decompressor
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)

    click.echo("- Download image-builder")
    path = f"{ib_path}.tar"
    with requests.get(
        f"https://github.com/kubernetes-sigs/image-builder/tarball/{image_builder_version}",
        stream=True,
    ) as r:
        r.raise_for_status()
        with open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                rv = dec.decompress(chunk)
                if rv:
                    f.write(rv)

    click.echo("- Extract image-builder")
    with tarfile.open(path) as tar:
        members = []
        for member in tar.getmembers():
            p = Path(member.path)
            member.path = p.relative_to(*p.parts[:1])
            members.append(member)

        tar.extractall(ib_path, members=members)

    click.echo("- Create customization file")
    kubernetes_series = ".".join(version.split(".")[0:2])
    customization = {
        "kubernetes_deb_version": f"{version.replace('v', '')}-1.1",
        "kubernetes_rpm_version": f"{version.replace('v', '')}",
        "kubernetes_semver": f"{version}",
        "kubernetes_series": f"{kubernetes_series}",
        # https://github.com/flatcar/Flatcar/issues/823
        "ansible_user_vars": f"oem_id=openstack {extra_ansible_user_vars}",
        "node_custom_roles_pre": f"{node_custom_roles_pre}",
    }

    # NOTE(mnaser): We use the latest tested daily ISO for Ubuntu 22.04 in order
    #               to avoid a lengthy upgrade process.

    # NOTE(jrosser): This can be uncommented again when
    # https://github.com/vexxhost/magnum-cluster-api/issues/378 is fixed
    # if operating_system == "ubuntu-2204":
    #    iso = "jammy-live-server-amd64.iso"
    #
    #    customization["iso_url"] = (
    #        f"http://cdimage.ubuntu.com/ubuntu-server/jammy/daily-live/current/{iso}"
    #    )
    #
    #    # Get the SHA256 sum for the ISO
    #    r = requests.get(
    #        "http://cdimage.ubuntu.com/ubuntu-server/jammy/daily-live/current/SHA256SUMS"
    #    )
    #    r.raise_for_status()
    #    for line in r.text.splitlines():
    #        if iso in line:
    #            customization["iso_checksum"] = line.split()[0]
    #            break
    #
    #    # Assert that we have the checksum
    #    assert "iso_checksum" in customization

    if operating_system == "rockylinux-8":
        iso = "Rocky-x86_64-minimal.iso"

        customization["iso_url"] = (
            f"https://download.rockylinux.org/pub/rocky/8/isos/x86_64/{iso}"
        )

        # Get the SHA256 sum for the ISO
        r = requests.get(
            "https://download.rockylinux.org/pub/rocky/8/isos/x86_64/Rocky-x86_64-minimal.iso.CHECKSUM"
        )
        r.raise_for_status()
        for line in r.text.splitlines():
            if iso in line and "SHA256" in line:
                customization["iso_checksum"] = line.split()[3]
                break

        # Assert that we have the checksum
        assert "iso_checksum" in customization

    if operating_system == "rockylinux-9":
        iso = "Rocky-x86_64-minimal.iso"

        customization["iso_url"] = (
            f"https://download.rockylinux.org/pub/rocky/9/isos/x86_64/{iso}"
        )

        # Get the SHA256 sum for the ISO
        r = requests.get(
            "https://download.rockylinux.org/pub/rocky/9/isos/x86_64/Rocky-x86_64-minimal.iso.CHECKSUM"
        )
        r.raise_for_status()
        for line in r.text.splitlines():
            if iso in line and "SHA256" in line:
                customization["iso_checksum"] = line.split()[3]
                break

        # Assert that we have the checksum
        assert "iso_checksum" in customization

    # NOTE(mnaser): Let's set number of CPUs to equal the number of CPUs on the
    #               host to speed up the build process.
    customization["cpus"] = str(os.cpu_count())

    # NOTE(mnaser): We set the memory of the VM to 50% of the total memory
    #               of the system.
    with open("/proc/meminfo", "r") as f:
        for line in f:
            if line.startswith("MemTotal:"):
                total_memory_kb = int(line.split()[1])
                break
    total_memory_mb = total_memory_kb / 1024
    customization["memory"] = str(int(total_memory_mb * 0.5))

    with tempfile.NamedTemporaryFile(suffix=".json") as fp:
        fp.write(json.dumps(customization).encode("utf-8"))
        fp.flush()

        click.echo("- Build image")
        subprocess.run(
            [
                "/usr/bin/newgrp",
                "kvm",
            ],
            input=textwrap.dedent(
                f"""\
                /usr/bin/make \
                -C \
                {ib_path}/images/capi \
                build-qemu-{operating_system}
            """
            ).encode("utf-8"),
            env={
                **os.environ,
                **{
                    "PATH": f"{expanduser('~/.local/bin')}:{os.environ['PATH']}",
                    "PACKER_VAR_FILES": fp.name,
                },
            },
        )

    # Try and detect the target image path since it can be different depending
    # on the operating system, so we scan the output directory for the file
    # that matches the pattern.
    target = None
    for root, dirs, files in os.walk(output_path):
        if len(files) > 1:
            ctx.fail(f"Unexpected number of files in {root}")
        if files:
            target = f"{root}/{files[0]}"

    if target is None:
        ctx.fail("Unable to detect target image")

    # Copy from the target to the current working directory
    os.rename(target, f"{operating_system}-kube-{version}.qcow2")



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import os
import socket
from datetime import datetime, timezone
from pathlib import Path

import jinja2
import pykube
import yaml
from oslo_log import log as logging
from oslo_service import periodic_task

import magnum_cluster_api.privsep.haproxy
from magnum_cluster_api import clients, conf, objects
from magnum_cluster_api.proxy import structs, utils

CONF = conf.CONF
LOG = logging.getLogger(__name__)


class ProxyManager(periodic_task.PeriodicTasks):
    def __init__(self):
        super(ProxyManager, self).__init__(CONF)

        self.api = clients.get_pykube_api()
        self.checksum = None
        self.template = jinja2.Environment(
            loader=jinja2.PackageLoader("magnum_cluster_api.proxy"),
            autoescape=jinja2.select_autoescape(),
        ).get_template("haproxy.cfg.j2")
        self.haproxy_port = utils.find_free_port(
            port_hint=int(os.getenv("PROXY_PORT", 0))
        )
        self.haproxy_bind = os.getenv("PROXY_BIND", "*")
        self.haproxy_pid = None

    def periodic_tasks(self, context, raise_on_error=False):
        return self.run_periodic_tasks(context, raise_on_error=raise_on_error)

    def _sync_haproxy(self, proxied_clusters: list):
        # Generate HAproxy config
        config = self.template.render(
            pid_file=CONF.proxy.haproxy_pid_path,
            port=self.haproxy_port,
            bind=self.haproxy_bind,
            clusters=proxied_clusters,
        )

        # Skip if no change has been done
        if self.checksum == hash(config):
            return

        LOG.info("Detected configuration change, reloading HAproxy")

        cfg_file = Path.home() / ".magnum-cluster-api-proxy" / "haproxy.cfg"
        cfg_file.parent.mkdir(parents=True, exist_ok=True)
        cfg_file.write_text(config)

        if self.haproxy_pid is None:
            self.haproxy_pid = magnum_cluster_api.privsep.haproxy.start(str(cfg_file))
        else:
            magnum_cluster_api.privsep.haproxy.reload()

        # Update checksum
        self.checksum = hash(config)

    def _sync_services(self, proxied_clusters: list, openstack_clusters: list):
        labels = {
            structs.ProxiedCluster.SERVICE_LABEL: "true",
        }

        # Get all services
        services = pykube.Service.objects(self.api, namespace="magnum-system").filter(
            selector=labels
        )

        # Generate list of all cluster names
        cluster_names = [
            cluster.metadata["labels"]["cluster.x-k8s.io/cluster-name"]
            for cluster in openstack_clusters
        ]

        # Remove any services that are not supposed to be there
        for service in services:
            if service.name not in cluster_names:
                LOG.info(
                    "Deleting service %s since the cluster does not exist",
                    service.name,
                )
                service.delete()

        # Create a list of ports to expose
        ports = [
            {
                "name": "https",
                "port": 6443,
                "targetPort": 6443,
                "protocol": "TCP",
            },
        ]

        # Create any services that are missing
        for cluster in proxied_clusters:
            service_name = cluster.name

            try:
                service = pykube.Service.objects(
                    self.api, namespace="magnum-system"
                ).get(name=service_name)
            except pykube.exceptions.ObjectDoesNotExist:
                LOG.info(
                    "Creating service %s",
                    service_name,
                )
                pykube.objects.Service(
                    self.api,
                    {
                        "apiVersion": pykube.Service.version,
                        "kind": pykube.Service.kind,
                        "metadata": {
                            "name": service_name,
                            "namespace": "magnum-system",
                            "labels": labels,
                        },
                        "spec": {
                            "ports": ports,
                        },
                    },
                ).create()
                service = pykube.Service.objects(
                    self.api, namespace="magnum-system"
                ).get(name=service_name)

            if (
                service.metadata["labels"] != labels
                or service.obj["spec"]["ports"] != ports
            ):
                LOG.info("Updating service %s", service.name)
                service.metadata["labels"] = labels
                service.obj["spec"]["ports"] = ports
                service.update()

    def _sync_endpoint_slices(self, proxied_clusters: list):
        hostname = socket.gethostname()

        # Get list of all endpoint slices assigned to this host
        endpoint_slices_for_host = objects.EndpointSlice.objects(
            self.api, namespace="magnum-system"
        ).filter(selector={structs.ProxiedCluster.NODE_LABEL: hostname})

        # Get list of all endpoint slices that are supposed to exist
        proxied_clusters_endpoint_slice_names = [
            cluster.endpoint_slice_name for cluster in proxied_clusters
        ]

        # Delete EndpointSlices that are no longer needed or if they are not healthy
        for endpoint_slice in endpoint_slices_for_host:
            if endpoint_slice.name not in proxied_clusters_endpoint_slice_names:
                LOG.info(
                    "Deleting EndpointSlice %s since it is not proxied on this host",
                    endpoint_slice.name,
                )
                endpoint_slice.delete()
                continue

            proxied_cluster = structs.ProxiedCluster.from_endpoint_slice(endpoint_slice)
            if not proxied_cluster.healthy:
                LOG.info(
                    "Deleting unhealthy EndpointSlice %s since the backend is unhealthy",
                    endpoint_slice.name,
                )
                endpoint_slice.delete()
                continue

        # Create EndpointSlices for each cluster
        for proxied_cluster in proxied_clusters:
            # Skip creating endpoint slices if the backend is unhealthy
            if not proxied_cluster.healthy:
                continue

            endpoint_slice_ports = proxied_cluster.endpoint_slice_ports(
                haproxy_port=self.haproxy_port
            )

            try:
                endpoint_slice = objects.EndpointSlice.objects(
                    self.api, namespace="magnum-system"
                ).get(name=proxied_cluster.endpoint_slice_name)
            except pykube.exceptions.ObjectDoesNotExist:
                LOG.info(
                    "Creating EndpointSlice %s", proxied_cluster.endpoint_slice_name
                )
                objects.EndpointSlice(
                    self.api,
                    {
                        "apiVersion": objects.EndpointSlice.version,
                        "kind": objects.EndpointSlice.kind,
                        "metadata": {
                            "name": proxied_cluster.endpoint_slice_name,
                            "namespace": "magnum-system",
                            "labels": proxied_cluster.endpoint_slice_labels,
                            "annotations": proxied_cluster.endpoint_slice_annotations,
                        },
                        "addressType": "IPv4",
                        "endpoints": proxied_cluster.endpoint_slice_endpoints,
                        "ports": endpoint_slice_ports,
                    },
                ).create()
                endpoint_slice = objects.EndpointSlice.objects(
                    self.api, namespace="magnum-system"
                ).get(name=proxied_cluster.endpoint_slice_name)

            # NOTE(mnaser): We always update the annotations since it contains the timestamp
            #               which we need for liveness.
            endpoint_slice.metadata["annotations"] = (
                proxied_cluster.endpoint_slice_annotations
            )
            endpoint_slice.update()

            if (
                endpoint_slice.metadata["labels"]
                != proxied_cluster.endpoint_slice_labels
                or endpoint_slice.obj["endpoints"]
                != proxied_cluster.endpoint_slice_endpoints
                or endpoint_slice.obj["ports"] != endpoint_slice_ports
            ):
                if (
                    endpoint_slice.metadata["labels"]
                    != proxied_cluster.endpoint_slice_labels
                ):
                    LOG.info(
                        "old_labels: %s, new_labels: %s",
                        endpoint_slice.labels,
                        proxied_cluster.endpoint_slice_labels,
                    )
                    endpoint_slice.metadata["labels"] = (
                        proxied_cluster.endpoint_slice_labels
                    )

                if (
                    endpoint_slice.obj["endpoints"]
                    != proxied_cluster.endpoint_slice_endpoints
                ):
                    LOG.info(
                        "old_endpoints: %s, new_endpoints: %s",
                        endpoint_slice.obj["endpoints"],
                        proxied_cluster.endpoint_slice_endpoints,
                    )
                    endpoint_slice.obj["endpoints"] = (
                        proxied_cluster.endpoint_slice_endpoints
                    )

                if endpoint_slice.obj["ports"] != endpoint_slice_ports:
                    LOG.info(
                        "old_ports: %s, new_ports: %s",
                        endpoint_slice.obj["ports"],
                        endpoint_slice_ports,
                    )
                    endpoint_slice.obj["ports"] = endpoint_slice_ports

                LOG.info("Updating EndpointSlice %s", endpoint_slice.name)
                endpoint_slice.update()

    def _sync_kubeconfigs(self, proxied_clusters: list):
        for cluster in proxied_clusters:
            # NOTE(mnaser): We only modify the `kubeconfig` if the cluster does
            #               not have a floating IP enabled.
            endpoint = f"https://{cluster.name}.magnum-system:6443"

            # Get the kubeconfig secret
            try:
                secret = pykube.Secret.objects(self.api, namespace="magnum-system").get(
                    name=cluster.kubeconfig_secret_name
                )
            except pykube.exceptions.ObjectDoesNotExist:
                LOG.warning(
                    "Kubeconfig secret %s does not exist",
                    cluster.kubeconfig_secret_name,
                )
                return

            # Get the kubeconfig from the secret
            kubeconfig = base64.b64decode(secret.obj["data"]["value"]).decode("utf-8")
            kubeconfig = yaml.safe_load(kubeconfig)

            # Check if the kubeconfig needs to be updated
            if kubeconfig["clusters"][0]["cluster"]["server"] == endpoint:
                continue

            # Update the kubeconfig endpoint
            LOG.info("Updating kubeconfig %s", cluster.kubeconfig_secret_name)
            kubeconfig["clusters"][0]["cluster"]["server"] = endpoint

            # Update the secret with the new kubeconfig
            secret.obj["data"]["value"] = base64.b64encode(
                yaml.safe_dump(kubeconfig).encode("utf-8")
            ).decode("utf-8")
            secret.update()

    def _cleanup_endpoint_slices(self):
        # Get list of all endpoint slices managed by the proxy service
        endpoint_slices = objects.EndpointSlice.objects(
            self.api, namespace="magnum-system"
        ).filter(selector={structs.ProxiedCluster.SERVICE_LABEL: "true"})

        # Look if any of the endpoint slices should be expired (aka >30s age)
        for endpoint_slice in endpoint_slices:
            timestamp = datetime.fromisoformat(
                endpoint_slice.metadata["annotations"][
                    structs.ProxiedCluster.TIMESTAMP_ANNOTATION
                ]
            )

            if (datetime.now(timezone.utc) - timestamp).total_seconds() > 30:
                LOG.info("Deleting expired EndpointSlice %s", endpoint_slice.name)
                endpoint_slice.delete()

    @periodic_task.periodic_task(spacing=10, run_immediately=True)
    def sync(self, context):
        # Generate list of all clusters
        clusters = objects.OpenStackCluster.objects(
            self.api, namespace="magnum-system"
        ).all()

        # Generate list of proxied clusters
        proxied_clusters = []
        for cluster in clusters:
            proxied_cluster = structs.ProxiedCluster.from_openstack_cluster(cluster)
            if proxied_cluster:
                proxied_clusters.append(proxied_cluster)

        self._sync_haproxy(proxied_clusters)
        self._sync_services(proxied_clusters, clusters)
        self._sync_endpoint_slices(proxied_clusters)
        self._sync_kubeconfigs(proxied_clusters)
        self._cleanup_endpoint_slices()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import semver
from magnum import objects
from magnum.common import context, exception
from oslo_log import log as logging

from magnum_cluster_api import clients, utils

LOG = logging.getLogger(__name__)


def is_enabled(
    cluster: objects.Cluster,
    label_flag: str,
    service_name: str,
) -> bool:
    return utils.get_cluster_label_as_bool(
        cluster, label_flag, True
    ) and is_service_enabled(service_name)


def is_service_enabled(service_type: str) -> bool:
    """Check if service is deployed in the cloud."""

    admin_context = context.make_admin_context()
    osc = clients.get_openstack_api(admin_context)
    keystone = osc.keystone()

    try:
        service = keystone.client.services.list(type=service_type)
    except Exception:
        LOG.exception("Failed to list services")
        raise exception.ServicesListFailed()

    if service and service[0].enabled:
        return True

    LOG.info("There is no %s service enabled in the cloud.", service_type)
    return False


def get_cloud_provider_image(
    cluster: objects.Cluster, tag_label: str, image_name: str
) -> str:
    tag = get_cloud_provider_tag(cluster, tag_label)
    version = semver.VersionInfo.parse(tag[1:])

    repository = "registry.k8s.io/provider-os"
    if version.major == 1 and version.minor < 24:
        repository = "docker.io/k8scloudprovider"

    return f"{repository}/{image_name}:{tag}"


def get_cloud_provider_tag(cluster: objects.Cluster, label: str) -> str:
    tag_label = cluster.labels.get(label, None)
    if tag_label:
        return tag_label

    kube_tag = utils.get_kube_tag(cluster)
    version = semver.VersionInfo.parse(kube_tag[1:])

    tag = None
    if version.major == 1 and version.minor == 23:
        tag = "v1.23.4"
    elif version.major == 1 and version.minor == 24:
        tag = "v1.24.6"
    elif version.major == 1 and version.minor == 25:
        tag = "v1.25.6"
    elif version.major == 1 and version.minor == 26:
        tag = "v1.26.3"
    elif version.major == 1 and version.minor == 27:
        # TODO(mnaser): There is no 1.27 release yet, so we're using
        #               the latest 1.26 release for now.
        tag = "v1.26.3"

    if tag is None:
        raise ValueError(
            f"Unsupported Kubernetes version: {version}. "
            "Please specify a supported version in the cluster template."
        )

    return tag



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fcntl
import socket
import struct
from contextlib import closing


def get_default_gateway_interface():
    with open("/proc/net/route") as f:
        for line in f:
            fields = line.strip().split()
            if fields[1] != "00000000" or not int(fields[3], 16) & 2:
                continue
            return fields[0]


def get_default_ip_address():
    interface = get_default_gateway_interface()
    if interface:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        ip_address = socket.inet_ntoa(
            fcntl.ioctl(
                sock.fileno(),
                0x8915,  # SIOCGIFADDR
                struct.pack("256s", interface.encode("utf-8")[:15]),
            )[20:24]
        )
        return ip_address


def find_free_port(port_hint=0):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        try:
            s.bind(("", port_hint))
        except OSError:
            s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



# Copyright (c) 2024 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from datetime import datetime, timedelta
from unittest import mock

import openstack
import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from magnum.common import context as magnum_context  # type: ignore
from magnum.objects import fields
from magnum.tests.unit.objects import utils
from oslo_utils import uuidutils  # type: ignore
from responses import matchers

from magnum_cluster_api import driver


@pytest.fixture
def context():
    return magnum_context.RequestContext(
        auth_token_info={
            "token": {"project": {"id": "fake_project"}, "user": {"id": "fake_user"}}
        },
        project_id="fake_project",
        user_id="fake_user",
        is_admin=False,
    )


@pytest.fixture(scope="session")
def mock_cluster_lock(session_mocker):
    session_mocker.patch("kubernetes.config.load_config")
    session_mocker.patch("magnum_cluster_api.sync.ClusterLock.acquire")
    session_mocker.patch("magnum_cluster_api.sync.ClusterLock.release")


@pytest.fixture(scope="session")
def mock_validate_cluster(session_mocker):
    session_mocker.patch("magnum_cluster_api.utils.validate_cluster")


@pytest.fixture(scope="session")
def mock_validate_nodegroup(session_mocker):
    session_mocker.patch("magnum_cluster_api.utils.validate_nodegroup")


@pytest.fixture()
def ubuntu_driver(mock_cluster_lock):
    yield driver.UbuntuDriver()


@pytest.fixture(scope="session")
def image():
    return {
        "id": uuidutils.generate_uuid(),
        "os_distro": "ubuntu",
    }


@pytest.fixture(scope="session")
def mock_osc(session_mocker, image):
    mock_clients = session_mocker.patch(
        "magnum_cluster_api.clients.OpenStackClients"
    ).return_value

    # NOTE(mnaser): Since there are reference in the Magnum code to the
    #               OpenStackClients, we need to make sure we mock it
    #               in the Magnum code as well.
    session_mocker.patch(
        "magnum.common.clients.OpenStackClients",
        return_value=mock_clients,
    )

    # Keystone
    mock_keystone_client = mock_clients.keystone.return_value.client
    mock_keystone_client.application_credentials.create.return_value = (
        openstack.identity.v3.application_credential.ApplicationCredential(
            id="fake_id", secret="fake_secret"
        )
    )

    # Glance
    mock_glance_client = mock_clients.glance.return_value
    mock_glance_client.images.get.return_value = image

    # Cinder
    mock_cinder_client = mock_clients.cinder.return_value
    mock_cinder_client.volume_types.default.return_value.name = "__DEFAULT__"
    mock_clients.cinder_region_name.return_value = "RegionOne"

    # Others
    mock_clients.url_for.return_value = "http://fake_url"

    return mock_clients


@pytest.fixture(scope="session")
def mock_certificate() -> mock.MagicMock:
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    cert: x509.Certificate = (
        x509.CertificateBuilder()
        .subject_name(x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, "test")]))
        .issuer_name(x509.Name([x509.NameAttribute(x509.NameOID.COMMON_NAME, "test")]))
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now())
        .not_valid_after(datetime.now() + timedelta(days=365))
        .sign(key, hashes.SHA256(), default_backend())
    )

    fake_certificate = mock.MagicMock()
    fake_certificate.get_certificate.return_value = cert.public_bytes(
        serialization.Encoding.PEM
    )
    fake_certificate.get_private_key_passphrase.return_value = "fake_passphrase"
    fake_certificate.get_private_key.return_value = key.private_bytes(
        serialization.Encoding.PEM,
        serialization.PrivateFormat.PKCS8,
        serialization.BestAvailableEncryption(
            fake_certificate.get_private_key_passphrase.return_value.encode("utf-8")
        ),
    )

    return fake_certificate


@pytest.fixture(scope="session")
def mock_get_cluster_ca_certificate(session_mocker, mock_certificate) -> mock.MagicMock:
    mock_get_cluster_ca_certificate = session_mocker.patch(
        "magnum.conductor.handlers.common.cert_manager.get_cluster_ca_certificate"
    )
    mock_get_cluster_ca_certificate.return_value = mock_certificate

    return mock_get_cluster_ca_certificate


@pytest.fixture(scope="session")
def mock_get_cluster_magnum_cert(session_mocker, mock_certificate) -> mock.MagicMock:
    mock_get_cluster_ca_certificate = session_mocker.patch(
        "magnum.conductor.handlers.common.cert_manager.get_cluster_magnum_cert"
    )
    mock_get_cluster_ca_certificate.return_value = mock_certificate

    return mock_get_cluster_ca_certificate


@pytest.fixture(scope="session")
def mock_certificates(mock_get_cluster_ca_certificate, mock_get_cluster_magnum_cert):
    pass


@pytest.fixture(scope="session")
def kube_tag():
    return os.getenv("KUBE_TAG", "v1.25.3")


@pytest.fixture()
def cluster_template(context, image, kube_tag):
    cluster_template = utils.get_test_cluster_template(
        context,
        image_id=image["id"],
        master_lb_enabled=True,
        external_network_id=uuidutils.generate_uuid(),
        dns_nameserver="8.8.8.8",
        master_flavor_id=uuidutils.generate_uuid(),
        flavor_id=uuidutils.generate_uuid(),
        cluster_distro="ubuntu",
        labels={"kube_tag": kube_tag},
    )
    cluster_template.save = mock.MagicMock()

    return cluster_template


@pytest.fixture()
def control_plane_node_group_obj(context, cluster_template):
    node_group = utils.get_test_nodegroup(
        context,
        name="default-master",
        role="master",
        node_count=1,
        flavor_id=cluster_template.master_flavor_id,
        image_id=cluster_template.image_id,
        labels=cluster_template.labels,
        status=fields.ClusterStatus.CREATE_IN_PROGRESS,
    )
    node_group.save = mock.MagicMock()

    yield node_group


@pytest.fixture()
def worker_node_group_obj(context, cluster_template):
    node_group = utils.get_test_nodegroup(
        context,
        name="default-worker",
        role="worker",
        node_count=1,
        flavor_id=cluster_template.master_flavor_id,
        image_id=cluster_template.image_id,
        labels=cluster_template.labels,
        status=fields.ClusterStatus.CREATE_IN_PROGRESS,
    )
    node_group.save = mock.MagicMock()

    yield node_group


@pytest.fixture()
def cluster_obj(
    mocker, cluster_template, control_plane_node_group_obj, worker_node_group_obj
):
    mocker.patch(
        "magnum.objects.NodeGroup.list",
        return_value=[control_plane_node_group_obj, worker_node_group_obj],
    )

    mocker.patch(
        "magnum.objects.ClusterTemplate.get_by_uuid",
        return_value=cluster_template,
    )

    cluster = utils.get_test_cluster(
        context,
        keypair="fake_keypair",
        master_flavor_id=cluster_template.master_flavor_id,
        flavor_id=cluster_template.flavor_id,
        image_id=cluster_template.image_id,
        ca_cert_ref=uuidutils.generate_uuid(),
        magnum_cert_ref=uuidutils.generate_uuid(),
        etcd_ca_cert_ref=uuidutils.generate_uuid(),
        front_proxy_ca_cert_ref=uuidutils.generate_uuid(),
        labels=cluster_template.labels,
    )
    cluster.save = mock.MagicMock()

    return cluster


@pytest.fixture
def server_side_apply_matcher():
    return matchers.query_param_matcher(
        {
            "fieldManager": "atmosphere-operator",
            "force": "True",
        }
    )
